// DOS program to control WiNRADiO receivers.

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "wrapi.h"
#include "wrdef.h"

struct _SETTINGS {
	int		Port;
	int		Irq;

	DWORD	Frequency;
	int		Mode;
	int		Volume;
	BOOL	Mute;
	BOOL	Atten;
	BOOL	Power;
	int		Bfo;
	int		IfShift;
	BOOL	Agc;
	int		IfGain;

	BOOL	ShowLogo;
	BOOL	Quiet;
} Settings = {0, 0, 10000000L, RMD_AM, 5, FALSE, FALSE, TRUE, 0, 0, TRUE, 100, TRUE, FALSE};

long StrToInt(char *s, long min, long max, long def)
{
	long v = atol(s);

	if ((v >= min) && (v <= max))
		return v;
	else
		return def;
}
	

void ReadIniFile(char *SourcePath)
{
	char *p, s[128];
	FILE *IniFile;
	
	p = strrchr(strcpy(s, SourcePath), '\\');
	if (p) p[1] = 0;	// put NULL terminator after last backslash in path
	
	IniFile = fopen(strcat(s, "dosradio.ini"), "rt");

	if (IniFile)
	{
		BOOL InSection = FALSE;

		while (fgets(s, sizeof(s), IniFile))
		{
			p = strchr(s, '\n');
			if (p)
				*p = 0;		// eliminate EOL character

			if (!strlen(s))
				continue;	// ignore blank lines!

			if (s[0] == '[')
			{
				InSection = !stricmp(s, "[Main]");
				continue;
			}
			if (!InSection)
				continue;

			if ((p = strchr(s, '=')) == NULL)
				continue;

			*p = 0;
			// line = value, p+1 = data
			if (!stricmp(s, "Frequency"))
				Settings.Frequency = StrToInt(p + 1, 150000L, 1500000000L, 10000000L);
			else if (!stricmp(s, "Mode"))
				Settings.Mode = StrToInt(p + 1, RMD_CW, RMD_USB, RMD_AM);
			else if (!stricmp(s, "Volume"))
				Settings.Volume = StrToInt(p + 1, 0, 31, 5);
			else if (!stricmp(s, "Atten"))
				Settings.Atten = StrToInt(p + 1, 0, 1, 0);
			else if (!stricmp(s, "Mute"))
				Settings.Mute = StrToInt(p + 1, 0, 1, 0);
			else if (!stricmp(s, "Power"))
				Settings.Power = StrToInt(p + 1, 0, 1, 1);
			else if (!stricmp(s, "Bfo"))
				Settings.Bfo = StrToInt(p + 1, -3000, 3000, 0);
			else if (!stricmp(s, "IfShift"))
				Settings.IfShift = StrToInt(p + 1, -3000, 3000, 0);
			else if (!stricmp(s, "Agc"))
				Settings.Agc = StrToInt(p + 1, 0, 1, 1);
			else if (!stricmp(s, "IfGain"))
				Settings.IfGain = StrToInt(p + 1, 0, 100, 100);
			// else unrecognized!
		}
		fclose(IniFile);
	}
}

void WriteIniFile(char *SourcePath)
{
	char *p, s[128];
	FILE *IniFile;

	
	p = strrchr(strcpy(s, SourcePath), '\\');
	if (p) p[1] = 0;	// put NULL terminator after last backslash in path
	
	IniFile = fopen(strcat(s, "dosradio.ini"), "wt");

	if (IniFile)
	{
		fputs("[Main]\n", IniFile);
		fprintf(IniFile, "Frequency=%ld\n", Settings.Frequency);
		fprintf(IniFile, "Mode=%d\n", Settings.Mode);
		fprintf(IniFile, "Volume=%d\n", Settings.Volume);
		fprintf(IniFile, "Atten=%d\n", Settings.Atten);
		fprintf(IniFile, "Mute=%d\n", Settings.Mute);
		fprintf(IniFile, "Power=%d\n", Settings.Power);
		fprintf(IniFile, "BFO=%d\n", Settings.Bfo);
		fprintf(IniFile, "IFShift=%d\n", Settings.IfShift);
		fprintf(IniFile, "AGC=%d\n", Settings.Agc);
		fprintf(IniFile, "IFGain=%d\n", Settings.IfGain);
		fclose(IniFile);
	}
}

void ShowUsage(void)
{
	printf("Usage: DOSRADIO [commands]\n\n");
	printf("  Commands:      [] = optional, <> = required, {} = one of enclosed values\n\n");
	printf("    /P[ort]:<n>[,<irq>]   = port to use, 1-4 = serial port, 180-1b8 = internal\n");
	printf("    /F[req]:<n>[{k,M,G}][Hz] = frequency to set (default unit is MHz)\n");
	printf("    /M[ode]:<mode>        = mode to set, modes include:\n");
	printf("                              C[W],A[M],[FM]N,[FM]W,L[SB],U[SB],[FM]M,[FM]6\n");
	printf("    /V[ol]:<n>            = volume to set (from 0 to 31)\n");
	printf("    /D[x]                 = set attenuator to DX mode (off)\n");
	printf("    /L[ocal]              = set attenuator to Local mode (on)\n");
	printf("    /POW[er]:{ON,OFF}     = set power state (on or off)\n");
	printf("    /MU[te]:{ON,OFF}      = set mute state\n");
	printf("    /I[fshift]:<n>        = set IF shift value (not on WR-1000)\n");
	printf("    /B[fo]:<n>            = set BFO offset (on WR-1000)\n");
	printf("    /A[gc]:{ON,OFF}       = set AGC state (on WR-2000 and WR-3500+)\n");
	printf("    /G[ain]:<n>           = set manual IF gain (on WR-2000 and WR-3500+)\n");
	printf("    /N[ologo]             = suppress copyright message\n");
	printf("    /Q[uiet]              = do not display any messages\n");
	printf("    /?, /H[elp]           = show this help message\n");
}

void InvalidParameter(char *s)
{
	printf("Unrecogized parameter on the command line: \"%s\"\n\n", s);
	ShowUsage();
}

BOOL ProcessParameter(char *param)
{
	char *p, s[80];
	long v;
	double f;
	BOOL Valid = FALSE;

	strlwr(strcpy(s, param + 1));

	switch (*s)
	{
		case 'b':
			if ((p = strchr(s, ':')) == NULL)	// find colon separator
				break;	// no offset specified!
			if (strncmp(s, "bfo", p - s))
				break;	// the paramter was something other than a substring of "bfo"
			v = strtol(++p, &p, 10);
			if (*p)
				break;	// crap after bfo
			Settings.Bfo = v;
			Valid = TRUE;
			break;

		case 'd':	// DX mode (atten off)
			if (strstr("dx", s))	// make sure that "s" is "dx" or part of "dx"
			{
				Settings.Atten = FALSE;
				Valid = TRUE;
			}
			break;

		case 'f':	// frequency (can be in Hz, kHz, MHz or GHz (default is MHz)
			if ((p = strchr(s, ':')) == NULL)	// find colon separator
				break;	// no frequency specified!
			if (strncmp(s, "freq", p - s))
				break;	// the paramter was something other than a substring of "freq"
			f = strtod(++p, &p);
			switch (*p)
			{
				case '\0':	// default is MHz
					f *= 1e6;
					break;
				case 'h':
					if ((*++p) && (*p != 'z'))
					{
						InvalidParameter(param);
						return FALSE;
					}
					break;
				case 'g':
					f *= 1e3;
				case 'm':
					f *= 1e3;
				case 'k':
					f *= 1e3;	// check rest of string (make sure it is empty or "hz")
					p++;
					if (!(*p && strcmp("hz", p)))
						break;	// break if EOS or rest of string is "hz"
				default:
					InvalidParameter(param);
					return FALSE;
			}
			if ((f > 0) && (f < 2.1e9))
			{
				Settings.Frequency = (DWORD)(f + 0.5);
				Valid = TRUE;
			}
			break;

		case 'g':
			if ((p = strchr(s, ':')) == NULL)	// find colon separator
				break;	// no value specified!
			if (strncmp(s, "gain", p - s))
				break;	// the parameter was something other than a substring of "gain"
			v = strtol(++p, &p, 10);
			if (*p || (v < 0) || (v > 100))
				break;	// crap after value, or value out of range
			Settings.IfGain = v;
			Valid = TRUE;
			break;

		case 'h':
			if (strstr("help", s))
			{
				ShowUsage();
				return FALSE;
			}
			break;

		case 'i':
			if ((p = strchr(s, ':')) == NULL)	// find colon separator
				break;	// no shift specified!
			if (strncmp(s, "ifshift", p - s))
				break;	// the parameter was something other than a substring of "ifshift"
			v = strtol(++p, &p, 10);
			if (*p)
				break;	// crap after bfo
			Settings.IfShift = v;
			Valid = TRUE;
			break;

		case 'l':	// local attenuator
			if (strstr("local", s))	// make sure that "s" is "local" or part of "local"
			{
				Settings.Atten = TRUE;
				Valid = TRUE;
			}
			break;

		case 'm':	// mode or mute
			if ((p = strchr(s, ':')) == NULL)	// find colon separator
				break;

			if ((p - s > 1) && (!strncmp(s, "mute", p - s)))	// check for mute first
			{
				p++;	// p can be "on" or "off"
				if (!strcmp(p, "on"))
					Settings.Mute = TRUE;
				else if (!strcmp(p, "off"))
					Settings.Mute = FALSE;
				else
					break;
				Valid = TRUE;
				break;
			}
			// must be mode at this stage
			if (strncmp(s, "mode", p - s))
				break;	// was not mode string!!
			p++;	// skip colon

			if (!(strcmp(p, "cw") && strcmp(p, "c")))
				Settings.Mode = RMD_CW;
			else if (!(strcmp(p, "am") && strcmp(p, "a")))
				Settings.Mode = RMD_AM;
			else if (!(strcmp(p, "fmn") && strcmp(p, "n")))
				Settings.Mode = RMD_FMN;
			else if (!(strcmp(p, "fmw") && strcmp(p, "w")))
				Settings.Mode = RMD_FMW;
			else if (!(strcmp(p, "lsb") && strcmp(p, "l")))
				Settings.Mode = RMD_LSB;
			else if (!(strcmp(p, "usb") && strcmp(p, "u")))
				Settings.Mode = RMD_USB;
			else if (!(strcmp(p, "fmm") && strcmp(p, "m")))
				Settings.Mode = RMD_FMM;
			else if (!(strcmp(p, "fm6") && strcmp(p, "6")))
				Settings.Mode = RMD_FM6;
			else
				break;

			Valid = TRUE;
			break;

		case 'n':
			if (strstr("nologo", s))
			{
				Settings.ShowLogo = FALSE;
				Valid = TRUE;
			}
			break;

		case 'q':
			if (strstr("quiet", s))
			{
				Settings.Quiet = TRUE;
				Valid = TRUE;
			}
			break;

		case 'p':	// can be port or power (port to start with)
			if ((p = strchr(s, ':')) == NULL)	// find colon separator
				break;	// no value specified!
			if (strncmp(s, "port", p - s))
			{
				if (strncmp(s, "power", p - s))
					break;	// the paramter was something other than a substring of "port" or "power"
				p++;
				if (!strcmp(p, "on"))
					Settings.Power = TRUE;
				else if (!strcmp(p, "off"))
					Settings.Power = FALSE;
				else
					break;
				Valid = TRUE;
				break;
			}
			v = strtol(++p, &p, 16);
			if ((*p) && (*p != ','))
				break;	// crap after the value

			if ((v < 0) || ((v > 4) && (((v & 7) != 0) || (v < 0x180) || (v > 0x1b8))))
				break;	// invalid port specified
			Settings.Port = v;

			if (*p)
			{
				v = strtol(++p, &p, 10);
				if ((*p) || (v < 3) || (v > 15))
					break;	// crap after value (or value is invalid)!
				Settings.Irq = v;
			}
			Valid = TRUE;
			break;
					;
		case 'v':
			if ((p = strchr(s, ':')) == NULL)	// find colon separator
				break;	// no value specified!
			if (strncmp(s, "volume", p - s))
				break;	// the paramter was something other than a substring of "volume"
			v = strtol(++p, &p, 10);
			if (*p)
				break;	// crap after the volume
			Settings.Volume = v;
			Valid = TRUE;
			break;

		case '?':
			if (strlen(s) == 1)	// should only be a question mark!
			{
				ShowUsage();
				return FALSE;
			}
			break;

	}

	if (!Valid)
		InvalidParameter(param);	// show error message if not valid parameter

	return Valid;
}


int main(int argc, char *argv[])
{
	int hRadio, ss;

	// load DOSRADIO.INI if it exists
	ReadIniFile(argv[0]);
	
	while (argc-- > 1)
		// process each command line parameter
		if ((*argv[argc] == '-') || (*argv[argc] == '/'))
		{
			if (!ProcessParameter(argv[argc]))
				return -1;	// exit if returns true (either invalid parameter or help
		} 
		else
		{ 
			InvalidParameter(argv[argc]);
			return -1;
		}

	if (Settings.ShowLogo && !Settings.Quiet)
	{
		printf("WiNRADiO (R) for DOS, Version 2.50\n");
		printf("Copyright (C) WiNRADiO Communications, 1995-2000. All rights reserved\n\n");
	}

	if ((hRadio = OpenRadioDevice(Settings.Port, Settings.Irq)) == 0)
	{
		printf("Could not find a WiNRADiO receiver at port %x\n", Settings.Port);
		return -1;
	}

	if (!Settings.Quiet)
	{
		char szVer[32];
		RADIOINFO RadioInfo;

		RadioInfo.dwSize = sizeof(RADIOINFO);
		GetRadioDeviceInfo(hRadio, &RadioInfo);

		if (RadioInfo.wHWVer <= 0x10a)
			strcpy(szVer, "WR-1000");
		else
		{
			if ((RadioInfo.wHWVer >= 0x200) && (RadioInfo.wHWVer < 0x300))
				RadioInfo.wHWVer += 0x100;
			else
				if ((RadioInfo.wHWVer >= 0x300) && (RadioInfo.wHWVer < 0x400))
					RadioInfo.wHWVer -= 0x100;

			sprintf(szVer, "WR-%d%.2d0", RadioInfo.wHWVer >> 8, RadioInfo.wHWVer & 0xff);
		}
		if (RadioInfo.iHWInterface)
		   strcat(szVer, "e");
		else
		   strcat(szVer, "i");

		if (RadioInfo.dwFeatures & RIF_USVERSION)
			strcat(szVer, " (US version)");

		printf("A %s was found at port %x\n", szVer, RadioSettings[hRadio]->wIoAddr);
	}

	SetPower(hRadio, Settings.Power);
	SetMode(hRadio, Settings.Mode);
	SetFrequency(hRadio, Settings.Frequency);
	SetBFOOffset(hRadio, Settings.Bfo);
	SetIFShift(hRadio, Settings.IfShift);
	SetVolume(hRadio, Settings.Volume);
	SetMute(hRadio, Settings.Mute);
	SetAtten(hRadio, Settings.Atten);

	ss = GetSignalStrength(hRadio);

	WriteIniFile(argv[0]);

	CloseRadioDevice(hRadio);

	return ss;
}