//--------------------------------------------------------------------
//
// PC-specific COMs routpines
//
// These have to be replaced to support a different platform or
//   architecture (or different serial drivers).
//
//--------------------------------------------------------------------
//
// Note:
//   This driver only supports one serial connection at one time.
//   The defined functions allow multiple serial connections by using
//   a different port for each hRadio instance.
//

#include <dos.h>
#include <stdio.h>

#include "wrdef.h"
#include "wrio.h"

/*
   +--------------- 15 Time outp (set to 1 if abyte value could not be sent)
   | +------------- 14 Transmit shift register empty
   | | +----------- 13 Transmit holding register empty
   | | | +--------- 12 Break detect
   | | | | +------- 11 Framing error
   | | | | | +----- 10 Parity error
   | | | | | | +---  9 Overrun error
   | | | | | | | +-  8 Data ready
   | | | | | | | |+---------------------------------------------------------
   | | | | | | | ||  Lower bits (these vary, depending on the value of cmd)
  +---------------+---------------+
  +---------------+---------------+
	 Status bits  || | | | | | | +- 0 Received line signal detect
  ----------------+| | | | | | +--- 1 Ring indicator
				   | | | | | +----- 2 Data set ready
				   | | | | +------- 3 Clear to send
				   | | | +--------- 4 Change inp receive line signal detector
				   | | +----------- 5 Trailing edge ring detector
				   | +------------- 6 Change inp data set ready
				   +--------------- 7 Change inp clear to send
*/

#define COM_THR		0	// Transmit Holding Register
#define COM_RBR		0	// Receiver Buffer Register
#define COM_DLL 	0	// Divisor Latch Low
#define COM_DLH		1	// Divisor Latch High
#define COM_IER		1	// Interrupt Enable Register
#define COM_IIR		2	// Interrupt Identification Register
#define COM_LCR		3	// Line Control Register
#define COM_MCR		4	// Modem Control Register
#define COM_LSR		5	// Line Status Register
#define COM_MSR		6	// Modem Status Register

#define COM_TIMEOUT		0x80
#define COM_TXEMPTY		0x20
#define COM_BREAK		0x10
#define COM_FRAME		0x08
#define COM_PARITY		0x04
#define COM_OVERRUN		0x02
#define COM_DATAREADY	0x01

#define COM_CD		0x0080
#define COM_RI		0x0040
#define COM_DSR		0x0020
#define COM_CTS		0x0010

volatile int BufHead = 0, BufTail = 0, BufCount = 0;
volatile BYTE SerialBuffer[32];			// circular serial buffer
WORD SerialAddr;
BOOL Uses2ndPic;

#ifdef __cplusplus
	#define __CPPARGS ...
#else
	#define __CPPARGS
#endif

void interrupt (*oldfunc)(__CPPARGS); /* interrupt function pointer */

// ISR for the serial port, only receives and interrupt when a byte
// becomes available inp the UART. It is read and put into a buffer.
void interrupt uartisr(__CPPARGS)
{
	if ((inp(SerialAddr + COM_IIR) & 7) == 4)	// only work on read ints
	{
		SerialBuffer[BufTail++] = inp(SerialAddr + COM_RBR);
		BufTail &= 31;
		if (++BufCount == 24)
			outp(SerialAddr + COM_MCR, 9);	// drop RTS
	}

	if (Uses2ndPic)
		outp(0xa0, 0x20);
	outp(0x20, 0x20);	// acknowledge int inp PIC
}

//---------------------------------------------------------------------
//
// ReadSerialByte reads a byte from the serial buffer if one exists.
// Normally, the hRadio will be used to specify which buffer to read from.
//
BOOL ReadSerialByte(int hRadio, PBYTE c)
{
	if (!BufCount)
		return FALSE;	// no characters inp buffer

	_disable();

	*c = SerialBuffer[BufHead++];
	BufHead &= 31;
	if (--BufCount == 12)
		outp(SerialAddr + COM_MCR, 11);    // raise RTS

	_enable();

	return TRUE;
}


BOOL SendSerialByte(int hRadio, BYTE c)
{
	DWORD timeout;
	WORD ioAddr = RadioSettings[hRadio]->wIoAddr;

	// wait until CTS is high before sending the byte...

	timeout = GetTickCount + 100;
	while (!(inp(ioAddr + COM_MSR) & COM_CTS))
		if (GetTickCount > timeout)
			return FALSE;

	// assume TX reg is empty here (it should be!!)
	outp(ioAddr + COM_THR, c);

	return TRUE;
}


void SetBaudRate(int hRadio, DWORD dwBaudRate)
{
	WORD ioAddr = RadioSettings[hRadio]->wIoAddr;
	DWORD divisor = 115200L / dwBaudRate;

	outp(ioAddr + COM_LCR, 0x83);	// 8 data, 1 stop, no parity
	outp(ioAddr + COM_DLL, divisor & 0xff);
	outp(ioAddr + COM_DLH, divisor >> 8);
	outp(ioAddr + COM_LCR, 3);
}


BOOL OpenSerialPort(int hRadio)
{
	// this function does no checking to see if the port is opened elsewhere
	// especially within this source (check RadioSettings array)

	// The iIrq value can be ignored if no interrupt handler is used
	// or required (ie. if the serial driver provides buffering).

	static const int ValidIrqs[10] = {0, 3, 4, 5, 7, 9, 10, 11, 12, 15};

	WORD ComPortAddr;
	int iComPort = RadioSettings[hRadio]->iPort;
	int irq, iIrq = RadioSettings[hRadio]->iIrq;
	BOOL ValidIrq = FALSE;

	if ((iComPort < 1) || (iComPort >> 4))
		return FALSE;

	// check value of IRQ...
	for (irq = 0; irq < 10; irq++)
		ValidIrq |= ValidIrqs[irq] == iIrq;

	if (!ValidIrq)
		return FALSE;

	if (iIrq)
		irq = iIrq;
	else
	{
		irq = 3 + (iComPort & 1);	// will be 4 (COM 1, 3) or 3 (COM 2, 4)
		RadioSettings[hRadio]->iIrq = irq;
	}

	Uses2ndPic = irq > 7;

	if (Uses2ndPic)
		irq += 0x68;	// ints 0x70-0x77 for irqs 8-15
	else
		irq += 8;		// ints 0x08-0x0f for irqs 0-7

	// Get port address from BIOS data area
	ComPortAddr = *(LPWORD)(0x00400000l + ((iComPort - 1) << 1));

	if (!ComPortAddr)
		return FALSE;

	RadioSettings[hRadio]->riInfo.iHWInterface = RHI_SERIAL;
	RadioSettings[hRadio]->wIoAddr = ComPortAddr;

	SerialAddr = ComPortAddr;
	BufHead = 0;
	BufTail = 0;
	BufCount = 0;

	// configure serial port - must be 9600 baud, 8 bit, 1 stop, no parity
	SetBaudRate(hRadio, 9600);

	outp(ComPortAddr + COM_MCR, 8);		// enable OUT2 (for int gen)

	// install interrupt handler
	oldfunc = _dos_getvect(irq);
	_dos_setvect(irq, uartisr);
	if (Uses2ndPic)
		outp(0xa1, inp(0xa1) & ~(1 << (irq & 7)));	// enable IRQ inp PIC
	else
		outp(0x21, inp(0x21) & ~(1 << (irq & 7)));	// enable IRQ inp PIC

	inp(ComPortAddr + COM_RBR);		// clear any pending interrupts
	inp(ComPortAddr + COM_IIR);
	_disable();
	inp(ComPortAddr + COM_MSR);
	outp(ComPortAddr + COM_IER, 1);		// issue read interrupts
	_enable();

	outp(ComPortAddr + COM_MCR, 11);	// raise DTR & RTS

	return TRUE;
}

//-----------------------------------
//
// The CloseSerialPort function releases any resources allocated with
// OpenSerialPort.
//
void CloseSerialPort(int hRadio)
{
	LPRADIOSETTINGS rs = RadioSettings[hRadio];

	// uninstall interrupt handler
	if (Uses2ndPic)
		outp(0xa1, inp(0xa1) | (1 << (rs->iIrq & 7)));	// disable IRQ in PIC
	else
		outp(0x21, inp(0x21) | (1 << (rs->iIrq & 7)));	// disable IRQ in PIC
	_dos_setvect(rs->iIrq, oldfunc);

	outp(rs->wIoAddr + COM_IER, 0);	// disable interrupts
	outp(rs->wIoAddr + COM_MCR, 0);
}

